import { EnumType, FieldInfo, Type } from "../net/serialization";

const TypeDefaultValue = {
    [Type.Bool]: () => false,
    [Type.U8]: () => 0,
    [Type.U16]: () => 0,
    [Type.U32]: () => 0,
    [Type.U64]: () => 0,
    [Type.U128]: () => 0,
    [Type.U256]: () => 0,
    [Type.Usize]: () => 0,
    [Type.I8]: () => 0,
    [Type.I16]: () => 0,
    [Type.I32]: () => 0,
    [Type.I64]: () => 0,
    [Type.I128]: () => 0,
    [Type.I256]: () => 0,
    [Type.Isize]: () => 0,
    [Type.F32]: () => 0,
    [Type.F64]: () => 0,
    [Type.BigI]: () => 0,
    [Type.Str]: () => '',
    [Type.Arr]: () => [],
    [Type.Map]: () => new Map,
    [Type.Option]: () => void 0,
    [Type.Struct]: (v) => initialStruct(v)
}
/**
 * 默认约定，一个结构体第一个字段为主键，如果不是主键则 idOrSpecialKey 为{ 'xxxkey' :'123'} 形式传入主键
 * 结构体中包含另外一个子结构体则可以 {'anothorStruct' :{k1:'v1',k2:'v2'}}
 * 生成对象并赋默认值, 根据需要可自行调整
 * @param Struct 结构体
 * @param idOrSpecialKey 初始化数值
 */
export function initialStruct(fields: FieldInfo[]) {
    let temp = Object.create(null);
    for (let i = 0; i < fields.length; i++) {
        let field = fields[i], key = field.name;
        //类型是一个Struct
        if (field.ftype.type == Type.Struct) {
            temp[key] = initialStruct(field.ftype.structType.fields);
            continue;
        }
        temp[key] = TypeDefaultValue[field.ftype.type]()
    }
    return temp;
}

window['initialStruct'] = initialStruct;

//============================数据序列化========================================
export const serializeField = {
    [Type.U32]: (v) => v,
    [Type.Str]: (v) => v,
    [Type.Bool]: (v) => v,
    [Type.Arr]: (v) => JSON.stringify(v),
    [Type.Struct]: (v, t: EnumType) => serialize(v, t.structType.fields),
    [Type.Map]: (map) => {
        let mapList = [];
        for (let [k, v] of map) {
            mapList.push([k, v]);
        }
        return mapList;
    }
}
export const deserializeField = {
    [Type.Bool]: (v) => v === "true",
    [Type.Arr]: (v) => JSON.parse(v),
    [Type.Struct]: (v, t: EnumType) => deserialize(JSON.parse(v), t.structType.fields),
    [Type.Map]: (str, t: EnumType) => {//map只嵌套简单结构
        let [tk, tv] = t.mapType, temp = new Map;
        let mapList = JSON.parse(str);
        let keyType = Type[tk.type];
        //key是否是数字
        let isNum = checkNum(keyType);
        for (let i = 0; i < mapList.length; i++) {
            let [k, val] = mapList[i];
            //v是特殊结构
            if (tv.type == Type.Struct) {
                val = deserialize(val, tv.structType.fields);
            }
            else if (tv.type == Type.Map) {
                val = deserializeField[tv.type](val, tv);
            }
            temp.set(isNum ? +k : k, val);
        }
        return temp
    }
}
//数据json化
export const serialize = (data: any, fields: FieldInfo[]) => {
    let temp = Object.create(null);
    for (let field of fields) {
        let key = field.name, ftype = field.ftype;
        let originVal = data[key];
        let convertFn = serializeField[ftype.type];
        temp[key] = convertFn ? convertFn(originVal, ftype) : originVal;
    }
    return temp
}

//字符串数据化
export const deserialize = (source: any, fields: FieldInfo[]) => {
    let temp = Object.create(null);
    for (let field of fields) {
        let key = field.name, ftype = field.ftype;
        let originVal = source[key];
        let convertFn = deserializeField[ftype.type];
        let keyType = Type[ftype.type];
        let isNum = checkNum(keyType);
        temp[key] = convertFn ? convertFn(originVal, ftype) : isNum ? +originVal : originVal;
    }
    return temp
}
export const checkNum = (keyType: string) => keyType.startsWith("U") || keyType.startsWith("I") || keyType.startsWith("F") || keyType.startsWith("Big");
export const any2str = (data: any, beautify = false) => {
    if (typeof data == 'string') {
        return data;
    }
    else return JSON.stringify(data, null, beautify ? 4 : 0);
}