/**
 * 缓动函数:
 * Linear：无缓动效果；
 * 		+ 注: linear 只有一种效果匀速
 * Quadratic：二次方的缓动（t^2）；
 * Cubic：三次方的缓动（t^3）；
 * Quartic：四次方的缓动（t^4）；
 * Quintic：五次方的缓动（t^5）；
 * Sinusoidal：正弦曲线的缓动（sin(t)）；
 * Exponential：指数曲线的缓动（2^t）；
 * Circular：圆形曲线的缓动（sqrt(1-t^2)）；
 * Elastic：指数衰减的正弦曲线缓动；
 * Back：超过范围的三次方缓动（(s+1)*t^3 - s*t^2）；
 * Bounce：指数衰减的反弹缓动；
 * 
 * 每种缓动函数都由三种变化:
 * 
 * 1. In  加速
 * 2. Out 减速
 * 3. InOut  先加速后减速
 * 
 * 实现参考： http://easings.net/cn
 * 效果参考：https://developer.playcanvas.com/en/tutorials/tweening/
 */

// ============================== 导入

// ============================== 导出
/**
 * @description 计算缓动的当前值
 * @param t current time, 当前时间
 * @param b beginning value, 初始值 (开始的属性值)
 * @param e ending value, 结束值 (结束的属性值)
 * @param d duration, 持续时间 解释为:运动的总时间
 * @param func select rase function, 选择的缓动函数
 * @example
 * 		[0ms, 6ms]的时间内，以曲线linear的形式，从 5 移到 10；返回：2ms 时刻的数字
 * 			let position = calcTween(2, 5, 10, 6, linearTween);
 *      如果想要做完整插值，得写个定时器去推；
 */
export function calcTween(t: number, b: number, e: number, d: number, func: Function): number {
	return b + (e - b) * func(t / d);
}

/**
 * @description 无缓动效果；
 * @param k 插值点百分比位置，取值范围：[0, 1]
 * @returns 采样结果，值范围：[0, 1]
 */
export function linearTween(k: number): number {
	return k;
}

/**
 * @description 二次缓动（t^2）
 * @param k 插值点百分比位置，取值范围：[0, 1]
 * @returns 采样结果，值范围：[0, 1]
 */
export function quadInTween(k: number): number {
	return k * k;
}

/**
 * @description 二次缓动（t^2） 减速
 * @param k 插值点百分比位置，取值范围：[0, 1]
 * @returns 采样结果，值范围：[0, 1]
 */
export function quadOutTween(k: number): number {
	return k * (2 - k);
}

/**
 * @description 二次缓动（t^2） 先加速后减速
 * @param k 插值点百分比位置，取值范围：[0, 1]
 * @returns 采样结果，值范围：[0, 1]
 */
export function quadInOutTween(k: number): number {
	k *= 2;
	if (k < 1) {
		return k * k * 0.5;
	}

	return (--k * (k - 2) - 1) * - 0.5;
}

/**
 * @description 三次缓动（t^3）
 * @param k 插值点百分比位置，取值范围：[0, 1]
 * @returns 采样结果，值范围：[0, 1]
 */
export function cubicInTween(k: number): number {
	return k * k * k;
}

/**
 * @description 三次缓动（t^3） 减速
 * @param k 插值点百分比位置，取值范围：[0, 1]
 * @returns 采样结果，值范围：[0, 1]
 */
export function cubicOutTween(k: number): number {
	return --k * k * k + 1;
}

/**
 * @description 三次缓动（t^3） 先减速后加速
 * @param k 插值点百分比位置，取值范围：[0, 1]
 * @returns 采样结果，值范围：[0, 1]
 */
export function cubicInOutTween(k: number): number {
	k *= 2;
	if (k < 1) {
		return k * k * k * 0.5;
	}

	return ((k -= 2) * k * k + 2) * 0.5;
}

/**
 * @description 四次缓动（t^4）
 * @param k 插值点百分比位置，取值范围：[0, 1]
 * @returns 采样结果，值范围：[0, 1]
 */
export function quartInTween(k: number): number {
	return k * k * k * k;
}

/**
 * @description 四次缓动（t^4） 减速
 * @param k 插值点百分比位置，取值范围：[0, 1]
 * @returns 采样结果，值范围：[0, 1]
 */
export function quartOutTween(k: number): number {
	return 1 - (--k * k * k * k);
}

/**
 * @description 四次缓动（t^4） 先减速后加速
 * @param k 插值点百分比位置，取值范围：[0, 1]
 * @returns 采样结果，值范围：[0, 1]
 */
export function quartInOutTween(k: number): number {
	k *= 2;
	if (k < 1) {
		return k * k * k * k * 0.5;
	}

	return ((k -= 2) * k * k * k - 2) * - 0.5;
}

/**
 * @description 五次缓动（t^5）
 * @param k 插值点百分比位置，取值范围：[0, 1]
 * @returns 采样结果，值范围：[0, 1]
 */
export function quintInTween(k: number): number {
	return k * k * k * k * k;
}

/**
 * @description 五次缓动（t^5） 减速
 * @param k 插值点百分比位置，取值范围：[0, 1]
 * @returns 采样结果，值范围：[0, 1]
 */
export function quintOutTween(k: number): number {
	return --k * k * k * k * k + 1;
}

/**
 * @description 五次缓动（t^5） 先减速后加速
 * @param k 插值点百分比位置，取值范围：[0, 1]
 * @returns 采样结果，值范围：[0, 1]
 */
export function quintInOutTween(k: number): number {
	k *= 2;
	if (k < 1) {
		// tslint:disable:binary-expression-operand-order
		return 0.5 * k * k * k * k * k;
	}

	return ((k -= 2) * k * k * k * k + 2) * 0.5;
}

/**
 * @description sin缓动
 * @param k 插值点百分比位置，取值范围：[0, 1]
 * @returns 采样结果，值范围：[0, 1]
 */
export function sinInTween(k: number): number {
	return 1 - Math.cos(k * Math.PI / 2);
}

/**
 * @description sin缓动 减速
 * @param k 插值点百分比位置，取值范围：[0, 1]
 * @returns 采样结果，值范围：[0, 1]
 */
export function sinOutTween(k: number): number {
	return Math.sin(k * Math.PI / 2);
}

/**
 * @description sin缓动 先减速后加速
 * @param k 插值点百分比位置，取值范围：[0, 1]
 * @returns 采样结果，值范围：[0, 1]
 */
export function sinInOutTween(k: number): number {
	return 0.5 * (1 - Math.cos(Math.PI * k));
}

/**
 * @description exp缓动
 * @param k 插值点百分比位置，取值范围：[0, 1]
 * @returns 采样结果，值范围：[0, 1]
 */
export function expInTween(k: number): number {
	return k <= 0 ? 0 : Math.pow(1024, k - 1);
}

/**
 * @description exp缓动 减速
 * @param k 插值点百分比位置，取值范围：[0, 1]
 * @returns 采样结果，值范围：[0, 1]
 */
export function expOutTween(k: number): number {
	return k >= 1 ? 1 : 1 - Math.pow(2, - 10 * k);
}

/**
 * @description exp缓动 先减速后加速
 * @param k 插值点百分比位置，取值范围：[0, 1]
 * @returns 采样结果，值范围：[0, 1]
 */
export function expInOutTween(k: number): number {
	if (k <= 0) {
		return 0;
	}
	if (k >= 1) {
		return 1;
	}
	k *= 2;
	if (k < 1) {
		return 0.5 * Math.pow(1024, k - 1);
	}

	return 0.5 * (- Math.pow(2, - 10 * (k - 1)) + 2);
}

/**
 * @description circle缓动
 * @param k 插值点百分比位置，取值范围：[0, 1]
 * @returns 采样结果，值范围：[0, 1]
 */
export function circInTween(k: number): number {
	return 1 - Math.sqrt(1 - k * k);
}

/**
 * @description circle缓动 减速
 * @param k 插值点百分比位置，取值范围：[0, 1]
 * @returns 采样结果，值范围：[0, 1]
 */
export function circOutTween(k: number): number {
	return Math.sqrt(1 - (--k * k));
}

/**
 * @description circle缓动 先减速后加速
 * @param k 插值点百分比位置，取值范围：[0, 1]
 * @returns 采样结果，值范围：[0, 1]
 */
export function circInOutTween(k: number): number {
	k *= 2;
	if (k < 1) {
		return - 0.5 * (Math.sqrt(1 - k * k) - 1);
	}

	return 0.5 * (Math.sqrt(1 - (k -= 2) * k) + 1);
}

/**
 * @description 指数衰减的正弦曲线缓动
 * @param k 插值点百分比位置，取值范围：[0, 1]
 * @returns 采样结果
 */
export function elasticInTween(k: number): number {
	if (k <= 0) {
		return 0;
	}
	if (k >= 1) {
		return 1;
	}

	return -Math.pow(2, 10 * (k - 1)) * Math.sin((k - 1.1) * 5 * Math.PI);
}

/**
 * @description 指数衰减的正弦曲线缓动 减速
 * @param k 插值点百分比位置，取值范围：[0, 1]
 * @returns 采样结果
 */
export function elasticOutTween(k: number): number {
	if (k <= 0) {
		return 0;
	}
	if (k >= 1) {
		return 1;
	}

	return Math.pow(2, -10 * k) * Math.sin((k - 0.1) * 5 * Math.PI) + 1;
}

/**
 * @description 指数衰减的正弦曲线缓动 先减速后加速
 * @param k 插值点百分比位置，取值范围：[0, 1]
 * @returns 采样结果
 */
export function elasticInOutTween(k: number): number {
	if (k <= 0) {
		return 0;
	}
	if (k >= 1) {
		return 1;
	}
	k *= 2;
	if (k < 1) {
		return -0.5 * Math.pow(2, 10 * (k - 1)) * Math.sin((k - 1.1) * 5 * Math.PI);
	}

	return 0.5 * Math.pow(2, -10 * (k - 1)) * Math.sin((k - 1.1) * 5 * Math.PI) + 1;
}

/**
 * @description 超过范围的三次方缓动
 * @param k 插值点百分比位置，取值范围：[0, 1]
 * @returns 采样结果
 * @note 参考：https://github.com/sole/tween.js/blob/master/src/Tween.js
 */
export function backInTween(k: number): number {
	const s = 1.70158;

	return k * k * ((s + 1) * k - s);
}

/**
 * @description 超过范围的三次方缓动 减速
 * @param k 插值点百分比位置，取值范围：[0, 1]
 * @returns 采样结果
 * @note 参考：https://github.com/sole/tween.js/blob/master/src/Tween.js
 */
export function backOutTween(k: number): number {
	const s = 1.70158;

	return --k * k * ((s + 1) * k + s) + 1;
}

/**
 * @description 超过范围的三次方缓动 先减速后加速
 * @param k 插值点百分比位置，取值范围：[0, 1]
 * @returns 采样结果
 * @note 参考：https://github.com/sole/tween.js/blob/master/src/Tween.js
 */
export function backInOutTween(k: number): number {
	const s = 1.70158 * 1.525;
	k *= 2;
	if (k < 1) {
		return 0.5 * (k * k * ((s + 1) * k - s));
	}

	return 0.5 * ((k -= 2) * k * ((s + 1) * k + s) + 2);
}

/**
 * @description 指数衰减的反弹缓动
 * @param k 插值点百分比位置，取值范围：[0, 1]
 * @returns 采样结果，值范围：[0, 1]
 */
export function bounceInTween(k: number): number {
	return 1 - bounceOutTween(1 - k);
}

/**
 * @description 指数衰减的反弹缓动 减速
 * @param k 插值点百分比位置，取值范围：[0, 1]
 * @returns 采样结果，值范围：[0, 1]
 */
export function bounceOutTween(k: number): number {
	if (k < (1 / 2.75)) {
		return 7.5625 * k * k;
	}
	if (k < (2 / 2.75)) {
		return 7.5625 * (k -= (1.5 / 2.75)) * k + 0.75;
	}
	if (k < (2.5 / 2.75)) {
		return 7.5625 * (k -= (2.25 / 2.75)) * k + 0.9375;
	}

	return 7.5625 * (k -= (2.625 / 2.75)) * k + 0.984375;
}

/**
 * @description 指数衰减的反弹缓动 先减速后加速
 * @param k 插值点百分比位置，取值范围：[0, 1]
 * @returns 采样结果，值范围：[0, 1]
 */
export function bounceInOutTween(k: number): number {
	if (k < 0.5) {
		return (1 - bounceOutTween(1 - k * 2)) * 0.5;
	}

	return bounceOutTween(k * 2 - 1) * 0.5 + 0.5;
}

/**
 * 三次 二维平面 贝塞尔曲线
 * 可用 [cubic-bezier](https://cubic-bezier.com/) 工具 来 查看效果
 */
export class CubicBezier {
	px3: number;
	px2: number;
	px1: number;
	py1: number;
	py2: number;
	py3: number;
	epsilon: number

	/**
	 * @param a, b, c, d: cubic-bezier 的 四个值，取值范围其实不限；但是一般都建议 [0, 1]
	 */ 
	constructor(a, b, c, d) {

		this.px3 = 3 * a;
		this.px2 = 3 * (c - a) - this.px3;

		this.px1 = 1 - this.px3 - this.px2;
		this.py3 = 3 * b;

		this.py2 = 3 * (d - b) - this.py3;
		this.py1 = 1 - this.py3 - this.py2;
		
		// 目标精度
		this.epsilon = 1e-7;
	}

	/**
	 * 插值函数
	 * @param x 插值点百分比位置，取值范围：[0, 1]
	 * @returns 采样结果，值范围由a b c d决定，但一般值范围：[0, 1]
	 */
	solve(x) {
		if (x === 0 || x === 1) { // 对 0 和 1 两个特殊 t 不做计算
			return this.getY(x);
		}

		let t = x;

		for (let i = 0; i < 8; i++) { 
			// 进行 8 次迭代
			let g = this.getX(t) - x;
			
			// 检测误差到可以接受的范围
			if (Math.abs(g) < this.epsilon) { 
				return this.getY(t);
			}
			
			// 对 x 求导
			let d = (3 * this.px1 * t + 2 * this.px2) * t + this.px3; 
			if (Math.abs(d) < 1e-6) { 
				// 如果梯度过低，说明牛顿迭代法无法达到更高精度
				break
			}

			t = t - g / d;
		}
		// 对得到的近似 t 求 y
		return this.getY(t);
	}
	
	private getX(t) {
		return ((this.px1 * t + this.px2) * t + this.px3) * t;
	}

	private getY(t) {
		return ((this.py1 * t + this.py2) * t + this.py3) * t;
	}
}

// ============================== 本地
