import { HttpRequest, HttpResponse } from "pi_pt/init/httprpc";
import { get_pid, to_pid } from "pi_pt/util/util";
import { initProjectDef } from "pi_pt/init/load_project";
import { AuthDbNew, AuthDbKey, UserDb } from "management_page/server/db/user_db.struct";
import { Env } from "pi_ucore/db/env";
import * as User from "./user.event";
import { Session } from "pi_ucore/event/session_new";
import { toByteArray } from "pi_tools/base64";
import { randomString } from "pi_tools/util";
import { checkAuth } from "./user.event";
import { AUTH, FM, RESULT_NUM } from "../../../constant";
import { getPlugin } from "../core/plugins_mgr";
import { sendRpcPid } from "../core/pid_event";
import { resultErr, resultOk } from "../result_handle";

declare var env: Env;

// #[httpRpc=/mgr/api/porxy/login]
export const login = async (req: HttpRequest, resp: HttpResponse) => {
    let msg = req.getData();

    let data = new UserDb();
    data.user = msg.user;
    data.pwd = msg.pwd;
    let localEnv = getEnv(msg.user);
    let r = await User.login(data, localEnv);
    console.log("login!!!!!!!!!!! r:", r);
    if (r.status === "success") {
        let token = randomString(6);
        // 设置会话信息
        sessionMgr.setSession(msg.user, { user: msg.user, session: localEnv.get("session"), token: { token, create_time: Date.now() } })
        r.data.token = token;
        r.data.user = msg.user;
        return r;
    }
    return r;
}

// http授权转mqtt授权
// #[rpc=rpcServer, src=*]
export const loginMqtt = async (authorization: string, env: Env) => {
    const userData = checkHttpPortAuthHandle(authorization);
    // 替换session
    env.set("session", userData.session);
    return "success";

}

// #[httpRpc=/mgr/api/porxy/createUser]
export const createUser = async (req: HttpRequest, resp: HttpResponse) => {
    let userData = checkHttpPortAuth(req);
    let msg = req.getData();
    let localEnv = getEnv(userData.user);
    let data = new UserDb();
    data.user = msg.user;
    data.pwd = msg.pwd;
    return await User.createUser(data, localEnv);
}

// #[httpRpc=/mgr/api/porxy/delUser]
export const delUser = async (req: HttpRequest, resp: HttpResponse) => {
    let userData = checkHttpPortAuth(req);
    let msg = req.getData();
    let localEnv = getEnv(userData.user);
    return await User.delUser(msg.user, localEnv);
}

// #[httpRpc=/mgr/api/porxy/mdfPwd]
export const mdfPwd = async (req: HttpRequest, resp: HttpResponse) => {
    let userData = checkHttpPortAuth(req);
    let msg = req.getData();
    let localEnv = getEnv(userData.user);
    let data = new UserDb();
    data.user = msg.user;
    data.pwd = msg.pwd;
    return await User.mdfPwd(data, localEnv);
}

// #[httpRpc=/mgr/api/porxy/users]
export const users = async (req: HttpRequest, resp: HttpResponse) => {
    let userData = checkHttpPortAuth(req);
    let localEnv = getEnv(userData.user);
    return await User.users(undefined, localEnv);
}

// #[httpRpc=/mgr/api/porxy/setAuth]
export const setAuth = async (req: HttpRequest, resp: HttpResponse) => {
    let userData = checkHttpPortAuth(req);
    let msg = req.getData();
    let json = JSON.parse(msg[""]);

    console.log("setAuth!!!!!!!!!!!!", json);
    let localEnv = getEnv(userData.user);
    let auth = new AuthDbNew();
    auth.user = json.user;
    auth.auth = json.auth;
    return await User.setAuth(auth, localEnv);
}

// 管理端插件代理 body: {event: string, data: any}
// #[httpRpc=/mgr/api/porxy/plugins_rpc]
export const plugins = async (req: HttpRequest, resp: HttpResponse) => {
    let userData = checkHttpPortAuth(req);
    let msg = req.getData();
    let json = JSON.parse(msg[""]);
    const event = json.event;
    const data = json.data;
    const timeout = json.timeout;
    const gray = json.gray;
    let localEnv = getEnv(userData.user);

    const pluginData = await getPlugin(event);
    if (!pluginData) throw "插件不存在";
    // 权限字符串R、W、RW转换为AUTH枚举
    switch (pluginData.auth.toLowerCase()) {
        case "r":
            if (!checkAuth(pluginData.plugin_name, AUTH.R, localEnv)) throw "权限错误";
            break;
        case "w":
            if (!checkAuth(pluginData.plugin_name, AUTH.W, localEnv)) throw "权限错误";
            break;
        case "rw":
            if (!checkAuth(pluginData.plugin_name, AUTH.RW, localEnv)) throw "权限错误";
            break;
        default:
            throw "权限为空";
    }
    try {
        return resultOk(await sendRpcPid(event, data, timeout, gray));
    } catch (error) {
        console.log("plugins error:", error.message, error, error.stack);
        return resultErr(error.message);
    }
}




// 获取env
export const getEnv = (user: string): Env => {
    let localEnv = new Env();
    localEnv.dbMgr = env.dbMgr;
    localEnv.other = new Map();
    let session = sessionMgr.getSession(user) ? sessionMgr.getSession(user).session : new Session(Date.now().toString(), new Map(), () => { });

    localEnv.set("session", session);
    return localEnv;
}


const mapToJson = (map: Map<string, any>) => {
    let json = {};
    map.forEach((v, k) => {
        json[k] = v;
    })
    return json;
}

const jsonToMap = (json: object) => {
    let map = new Map();
    for (let k of Object.keys(json)) {
        console.log("jsonToMap:", { k, v: json[k] });
        map.set(k, json[k]);
    }
    return map;
}

const setCookie = (resp: HttpResponse, cookie: string) => {
    resp.setHeader("Set-Cookie", `attr=${cookie}`);
}

const getCookie = (req: HttpRequest): string => {
    let cookie = req.getHeader("cookie");
    return cookie
}

type UserData = {
    user: string,
    session: Session,
    token: {
        token: string,
        create_time: number
    },
}

class SessionMgr {
    sessions: Map<string, UserData>;
    constructor() {
        this.sessions = new Map();
    }

    setSession(user: string, userData: UserData) {
        this.sessions.set(user, userData);
    }

    getSession(user: string) {
        return this.sessions.get(user);
    }
}

let sessionMgr = new SessionMgr();


// httpPort权限检测
export const checkHttpPortAuth = (req: HttpRequest): UserData => {
    let auth = req.getHeader("authorization");
    // Basic dGVzdDoxMjM0NTY=
    return checkHttpPortAuthHandle(auth);
}

// httpPort权限检测
const checkHttpPortAuthHandle = (authorization: string): UserData => {
    try {
        let txt = new TextDecoder().decode(toByteArray(authorization.split(' ')[1]));
        const [username, password] = txt.split(':');
        let userData = sessionMgr.getSession(username);
        if (!userData) throw new Error("no user");
        if (password !== userData.token.token || userData.token.create_time + 7 * 24 * 60 * 60 * 1000 < Date.now()) {
            throw new Error("check auth error");
        };
        return userData;
    } catch (error) {
        console.log("checkHttpPortAuth auth error:", error.message, error.stack);
        throw new Error("auth error");
    }
}